package KL::App;

use strict;

use File::Basename qw(basename);
use File::Spec;
use File::Temp;
use File::Path qw(make_path rmtree);
use File::Copy qw(copy);
use Cwd qw(getcwd);

use KL::PlaformSpecificUtils;
use KL::AppBase;
use KL::IniFile;
use KL::KlNagentRelations;

use POSIX qw(setlocale);
BEGIN
{
    my $lang = substr(setlocale("LC_ALL"), 0, 2);
    my @LFS_LANGS = split(/\s*,\s*/, "en");
    if (grep {$_ eq $lang} @LFS_LANGS)
    {
        eval "require KL::Localization::$lang; KL::Localization::$lang->import();";
    }
    else
    {
        require KL::Localization::en;
        KL::Localization::en->import();
    }
}

my $APP_ID = 1046;

my $TASK_ID_UPDATER = KL::AppInfo::LFS_TASK_ID_UPDATE;
my $TASK_ID_OAS     = KL::AppInfo::LFS_TASK_ID_OAS;

my $KSN_ENABLED_SETTING = 'KSN_ENABLED';

my $BL_FILEPATH='/opt/kaspersky/kesl/libexec/kesl';
my $CONTROL_FILEPATH='/opt/kaspersky/kesl/bin/kesl-control';
my $SUPERVISOR_FILEPATH='/opt/kaspersky/kesl/shared/kesl';
my $GUI = '/opt/kaspersky/kesl/libexec/kesl-gui';
my $LENA_AUSHAPE_PLUGIN = '';
my $LENA_AUSHAPE = '';
my $LENA_LIBAUSHAPE = '';
my $LFS_FONTS_CONF = '/opt/kaspersky/kesl/resource/fonts.conf';
my $LFS_PRODUCT_NAME = 'kesl';
my $SELINUX_POLICY_DIR='/opt/kaspersky/kesl/shared/selinux/';
my $LFS_TEMP_DIRECTORY='/var/opt/kaspersky/kesl/common/temp/';

my $RC_NO_BASES = 66;
my $RC_RESTART_REQUIRED = 74;
my $RC_REBOOT_REQUIRED = 75;

use base 'KL::PlaformSpecificUtils';
use base 'KL::KlNagentRelations';

sub new
{
    my $class = shift;
    my $self = $class->SUPER::new($APP_ID);

    push @{$self->{'COMMANDS'}}, {
        'id' => 'help',
        'spec' => 'help',
        'sub' => 'do_show_help',
        'default' => undef,
        'descr' => $MSG_CMD_HELP_HELP,
    };

    $self->{'AUTOINSTALL'} = [
        {
            'id'        => 'EULA_AGREED',
            'list'      => [ 'Yes' ],
            'descr'     => $MSG_AI_HELP_EULA_AGREED,
        },
        {
            'id'        => 'PRIVACY_POLICY_AGREED',
            'list'      => [ 'Yes' ],
            'descr'     => $MSG_AI_HELP_PRIVACY_POLICY_AGREED,
        },
        {
            'id'        => 'USE_KSN',
            'list'      => [ 'Yes', 'No' ],
            'descr'     => $MSG_AI_HELP_USE_KSN,
        },
        {
            'id'        => 'LOCALE',
            'type'      => 'locale',
            'default'   => '',
            'descr'     => $MSG_AI_HELP_LOCALE,
        },
        {
            'id'        => 'INSTALL_LICENSE',
            'type'      => 'activation',
            'default'   => '',
            'descr'     => $MSG_AI_HELP_INSTALL_LICENSE,
        },
        {
            'id'        => 'UPDATER_SOURCE',
            'type'      => 'url_or_abspath',
            'list'      => [ 'KLServers', 'SCServer' ],
            'protocols' => [ 'https', 'http', 'ftp' ],
            'protocol_requred' => 0,
            'default'   => 'KLServers',
            'descr'     => $MSG_AI_HELP_UPDATER_SOURCE,
        },
        {
            'id'        => 'PROXY_SERVER',
            'type'      => 'url',
            'protocols' => [ 'http' ],
            'protocol_requred' => 1,
            'list'      => [ 'None' ],
            'default'   => 'None',
            'descr'     => $MSG_AI_HELP_PROXY_SERVER,
        },
        {
            'id'        => 'UPDATE_EXECUTE',
            'list'      => [ 'Yes', 'No' ],
            'default'   => 'Yes',
            'descr'     => $MSG_AI_HELP_UPDATE_EXECUTE,
        },
        {
            'id'        => 'KERNEL_SRCS_INSTALL',
            'list'      => [ 'Yes', 'No' ],
            'default'   => 'No',
            'descr'     => $MSG_AI_HELP_KERNEL_SRCS_INSTALL,
        },
        {
            'id'        => 'ADMIN_USER',
            'type'      => 'string',
            'default'   => '',
            'descr'     => $MSG_AI_HELP_ADMIN_USER,
        },
        {
            'id'        => 'CONFIGURE_SELINUX',
            'list'      => [ 'Yes', 'No' ],
            'default'   => 'Yes',
            'descr'     => $MSG_AI_HELP_CONFIGURE_SELINUX,
        },
    ];

    $self->load_settings();

    my $bin = sprintf(
        '%s/%s',
        $self->get_path('bins'),
        $self->appinfo('product', 'name'));

    $self->{'SETUP'} = $self->get_path('setup');
    $self->{'control'} = $self->get_path('control');
    $self->{'fwcontrol'} = $self->get_path('fwcontrol');

    return $self;
}

sub check_blocked_by_policy_result
{
    my $self = shift;
    my $ret  = shift;

    if ($ret == 68)
    {
        KL::Utils::warning($MSG_PRODUCT_IS_RUNNING_UNDER_POLICY);
        exit;
    }
}

sub check_invalid_permissions_result
{
    my $self = shift;
    my $ret = shift;
    my $out = shift;
    if ($ret == 76)
    {
        if (defined $out and length $out)
        {
            KL::Utils::fatal(5, $out);
        }
        else
        {
            KL::Utils::fatal(5, "Invalid permissions. Check /, /opt, /opt/kaspersky, /var, /var/opt, /var/opt/kaspersky, /usr, /usr/lib64. Only root user should have write access to these directories.");
        }
    }
}

sub _control
{
    my $self = shift;

    my $control = $self->{'control'};

    my @args = @_;
    foreach (@args) {
        $_ = KL::Utils::shell_escape($_);
    }

    KL::Utils::debug("_control: $control " . join(' ', @args));

    system($control, @args);

    my $ret = $?;
    if ($ret == -1) {
        KL::Utils::warning($MSG_CANT_EXECUTE, $control, join(' ', @args), $!);
        return -128;
    } elsif ($ret & 127) {
        $ret &= 127;
    } else {
        $ret >>= 8;
    }

    my $out = "";
    $self->check_blocked_by_policy_result($ret);
    $self->check_invalid_permissions_result($ret, $out);

    return $ret;
}

sub _control_nout
{
    my $self = shift;

    my $control = $self->{'control'};

    my ($ret, $out) = KL::Utils::execute($control, @_);
    KL::Utils::warning($MSG_CANT_EXECUTE, $control, join(' ', @_), $!) if $ret == -128;

    KL::Utils::debug("_control_nout( $control " . join(' ', @_) . " ) = $ret '$out'");

    $self->check_blocked_by_policy_result($ret);
    $self->check_invalid_permissions_result($ret, $out);

    return ($ret, $out);
}

sub _control_nerr
{
    my $self = shift;

    my $control = $self->{'control'};

    my ($ret, $out) = KL::Utils::execute_stderr($control, @_);
    KL::Utils::warning($MSG_CANT_EXECUTE, $control, join(' ', @_), $!) if $ret == -128;

    KL::Utils::debug("_control_nerr( $control " . join(' ', @_) . " ) = $ret '$out'");

    $self->check_blocked_by_policy_result($ret);
    $self->check_invalid_permissions_result($ret, $out);

    return ($ret, $out);
}

sub _get_settings_as_simple_hash
{
    my $self = shift;
    my $command = shift;
    my $taskid = shift;

    my ($ret, $str) = $self->_control_nout($command, $taskid, @_);
    if ($ret != 0) {
        KL::Utils::debug("get_task_settings($taskid) = $ret: $str");
        return ($ret, $str);
    }

    return ($ret, KL::Utils::parse_ini_content($str));
}

sub get_scheduler_settings_as_simple_hash
{
    my $self = shift;
    my $taskid = shift;

    return $self->_get_settings_as_simple_hash('--get-schedule', $taskid, @_);
}

sub get_task_settings_as_simple_hash
{
    my $self = shift;
    my $taskid = shift;

    return $self->_get_settings_as_simple_hash('--get-settings', $taskid, @_);
}

sub get_app_settings_as_simple_hash
{
    my $self = shift;

    return $self->_get_settings_as_simple_hash('--get-app-settings', '', @_);
}

sub get_task_settings
{
    my $self = shift;
    my $taskid = shift;

    my ($ret, $str) = $self->_control_nout('--get-settings', $taskid, @_);
    if ($ret != 0) {
        KL::Utils::debug("get_task_settings($taskid) = $ret: $str");
        return undef;
    }

    my $values = {};

    my @lines = split(/\n/, $str);

    foreach my $line (@lines) {
        if ($line =~ /^(.*?)=(.*)$/) {
            my $name = $1;
            my $value = $2;

            if ($value =~ /^"(.*)"$/) {
                $value = $1;
            }

            $values->{$name} = $value;
        }
    }

    return $values;
}

sub set_task_settings
{
    my $self = shift;
    my $taskid = shift;

    my @options;
    while ( @_ ) {
        my $name = shift;
        my $value = shift;
        push @options, "$name=$value";
    }

    my ($ret, $str) = $self->_control_nout('--set-settings', $taskid, @options);

    KL::Utils::debug("set_task_settings() = $ret: $str");

    return $ret == 0;
}

sub _refresh_settings
{
    my $self = shift;
    my $command = shift;
    my $taskid = shift;
    my $opts = shift;

    my $template = sprintf('%s_%d_XXXXX', $self->appinfo('product', 'name'), $taskid);
    my (undef, $filename) = File::Temp::tempfile($template, OPEN => 0,
        SUFFIX => '.ini', DIR => File::Spec->tmpdir());

    while ( @_ ) {
        my $section = '';
        my $name = shift;
        my $value = shift;

        if ($name =~ /^(.*)\.(.*?)$/) {
            $section = $1;
            $name = $2;
        }

        $opts->{$section}->{$name} = $value;
    }

    if (!KL::Utils::save_ini_file($filename, $opts)) {
        KL::Utils::debug("refresh_task_settings() cant update $filename");

        unlink($filename) if -f $filename;
        return (255, $MSG_CANNOT_SAVE_TEMP_FILE);
    }

    my ($ret, $str) = $self->_control_nout($command, $taskid,
        '--file', $filename);

    KL::Utils::debug("refresh_task_settings() set-settings $taskid = $ret: $str");

    if ($ret != 0) {
        unlink($filename) if -f $filename and not KL::Utils::debug_mode();

        return ($ret, $str);
    }

    unlink($filename) if -f $filename;

    return ($ret, $str);
}

sub refresh_scheduler_settings
{

    my $self = shift;
    my $taskid = shift;

    my ($res, $opts) = $self->get_scheduler_settings_as_simple_hash($taskid);
    if ($res != 0 || ! defined $opts)
    {
        return 0;
    }

    my ($ret, $msg) = $self->_refresh_settings('--set-schedule', $taskid, $opts, @_);
    return ($ret == 0 ? 1 : 0, $msg);
}

sub refresh_app_settings
{
    my $self = shift;
    my ($res, $opts) = $self->get_app_settings_as_simple_hash();
    if ($res != 0)
    {
        return ($res, $opts);
    }
    if (! defined $opts)
    {
        return (64, $opts);
    }
    my ($ret, $msg)  = $self->_refresh_settings('--set-app-settings', '', $opts, @_);
    return ($ret, $msg);
}

sub refresh_task_settings
{
    my $self = shift;
    my $taskid = shift;

    my ($res, $opts) = $self->get_task_settings_as_simple_hash($taskid);
    if ($res != 0)
    {
        return $res;
    }
    if (!defined $opts)
    {
        return 128;
    }
    ($res, $opts) = $self->_refresh_settings('--set-settings', $taskid, $opts, @_);
    return $res;
}


sub install_activation_impl
{
    my $self = shift;
    my $activation = shift;

    my ($ret, $output) = $self->_control_nout('-L', '--add-active-key', $activation);

    KL::Utils::debug("install_activation($activation) = $ret: $output");

    return ($ret, $output);
}

sub get_http_proxy
{
    my $self = shift;
    my ($res, $settings) = $self->get_app_settings_as_simple_hash();

    if($res != 0)
    {
        return ($res, $settings);
    }
    if (!defined $settings)
    {
        return (128, $MSG_GET_PROXY_FAILED);
    }

    $settings = $settings->{''};

    my $use_proxy = $settings->{'UseProxy'};

    if (lc($use_proxy) eq 'no')
    {
        return (0, 'n');
    }

    my $proxy_server = $settings->{'ProxyServer'};

    return (0, $proxy_server);
}

sub has_controls
{
    my $value = shift;
    return $value =~ m/[\x{00}-\x{1f}\x{7f}]/g;
}

sub check_rfc3986_authority_parts
{
    my $value = shift;
    return 0 if ($value =~ m/[^a-zA-Z\d\-\.\_\~\!\$\&\'\(\)\*\+\,\;\=\%]/g);
    return ! ($value =~ m/%[^\da-fA-F]|%.[^a-fA-F\d]|%.?$/g);
}

sub set_http_proxy
{
    my $self = shift;
    my $value = shift;

    my @options;

    if (has_controls($value))
    {
        return (0, KL::Utils::translate_string($MSG_FORBIDDEN_CONTROL_SEQUENCE, $value))
    }

    if ($value =~ /^no(ne)?$/i) {
        push  @options, 'UseProxy', 'no';
    }
    elsif ($value =~ /^(http:\/\/)?([^\/?#]*)?/)
    {
        $value = $2;

        if($value =~ /^((.*?)(:(.*?))?@)?(.*):(\d+)/)
        {
            my ($user, $pass, $host, $port) = ($2, $4,  $5, $6);

            if (check_rfc3986_authority_parts($user) == 0 ||
                check_rfc3986_authority_parts($pass) == 0 ||
                check_rfc3986_authority_parts($host) == 0)
            {
                return (2, KL::Utils::translate_string($MSG_USE_CODES_FOR_RFC3986_GEN_DELIMS_CHARACTERS))
            }

            if (!defined $host)
            {
                return (2, KL::Utils::translate_string($MSG_INCORRECT_PROXY_VALUE, $value));
            }
            push  @options, 'UseProxy', 'Yes';
            push  @options, 'ProxyServer', $value;
        }
        else
        {
            return (2, KL::Utils::translate_string($MSG_INCORRECT_PROXY_VALUE, $value));
        }
    }
    else
    {
        return (2, KL::Utils::translate_string($MSG_INCORRECT_PROXY_VALUE, $value));
    }

    my ($res, $message) = $self->refresh_app_settings(@options);

    return ($res, $message);
}

sub has_matched_line
{
    my $self = shift;
    my $strings = shift;
    my $regexp_string = shift;
    foreach my $line (split(/\n/, $strings)) {
        return 1 if ($line =~ m/$regexp_string/i);
    }
    return 0;
}

sub has_bases
{
    my $self = shift;
    my ($ret, $str) = $self->_control_nout('--app-info');

    if ($ret == 0 || $ret == $RC_RESTART_REQUIRED || $ret == $RC_REBOOT_REQUIRED)
    {
        return (0, 1);
    }
    elsif ($ret == $RC_NO_BASES)
    {
    	return (0, 0);
    }
    else
    {
        return ($ret, $str);
    }
}

sub run_updater
{
    my $self = shift;

    my $ret;
    my $updateEvents;
    if ($self->ai_mode)
    {
        ($ret, $updateEvents) =
            $self->_control_nout('--start-task', $TASK_ID_UPDATER, '-W');

        if ($self->has_matched_line($updateEvents, 'EventType=UpdateError'))
        {
            $ret = 65;
        }
    }
    else
    {
        ($ret, $updateEvents) =
            $self->_control_nerr('--start-task', $TASK_ID_UPDATER, '--progress');
    }

    print "$updateEvents" if $ret != 0;

    KL::Utils::debug("run_updater() = $ret: $updateEvents");

    return $ret;
}

sub set_updater_source
{
    my $self = shift;
    my $source = shift;

    my @options;

    if (!defined $source) {
        return (128, KL::Utils::translate_string($MSG_UPDATER_SOURCE_CHANGE_FAILED, $source));
    }

    if ($source =~ /KLServers/i) {
        push @options, 'SourceType', '"KLServers"';
    } elsif ($source =~ /SCServer/i) {
        push @options, 'SourceType', '"SCServer"';
    } else {
        push @options, 'CustomSources.item_0000.URL', '"' . $source . '"';
        push @options, 'CustomSources.item_0000.Enabled', 'Yes';
        push @options, 'SourceType', '"Custom"';
        push @options, 'UseKLServersWhenUnavailable', 'No';
    }

    my $res = $self->refresh_task_settings($TASK_ID_UPDATER, @options);
    if ($res == 0)
    {
        return ($res, KL::Utils::translate_string($MSG_UPDATER_SOURCE_CHANGED, $source));
    }

    return ($res, KL::Utils::translate_string($MSG_UPDATER_SOURCE_CHANGE_FAILED, $source));
}

sub set_autoupdates
{
    my $self = shift;
    my $enable = shift;

    my @current_options;
    if (lc($enable) eq 'yes')
    {
        push @current_options, 'RuleType', 'Minutely';
        push @current_options, 'StartTime', '00:00:00;60';
        push @current_options, 'RandomInterval', '99';
        push @current_options, 'RunMissedStartRules', 'Yes';
    }
    else
    {
        push @current_options, 'RuleType', 'Manual';
    }

    my ($ret, $message) = $self->refresh_scheduler_settings($TASK_ID_UPDATER, @current_options);

    if (not $ret)
    {
        return ($ret, $message);
    }
    if (lc($enable) eq 'yes')
    {
        return (1, KL::Utils::translate_string($MSG_AUTOUPDATE_HOURLY));
    }
    return (1, KL::Utils::translate_string($MSG_AUTOUPDATE_MANUALLY));
}

sub ai_validate_kernel_srcs
{
    my $self = shift;
    my $value = shift;

    if ($^O !~ /linux/i) {
        return 0;
    }

    if (-d $value && -f "$value/Makefile") {
        return 1;
    }

    return 0;
}

sub find_kernel_sources
{
    my $self = shift;

    my $os_name = KL::Utils::get_os_name();
    my $os_rev  = `uname -r`;
    $os_rev =~ s/\n//;

    my $kernel_srcs = "/lib/modules/$os_rev/build";

    if (!-d "$kernel_srcs") {
        if (-d "/usr/src/kernel-headers-$os_rev") {
                $kernel_srcs = "/usr/src/kernel-headers-$os_rev";
        } elsif (-d "/usr/src/linux" &&
                    ($os_name =~ /^ALT / ||
                     $os_name =~ /Debian/)) {
            $kernel_srcs = "/usr/src/linux";
        }
    }

    if ($self->ai_validate_kernel_srcs($kernel_srcs)) {
        return $kernel_srcs;
    }

    return undef;
}

sub check_fanotify_available
{
    my $self = shift;

    my $fanotify_checker = $self->get_path('fanotify-checker');
    if (!(-f $fanotify_checker) || KL::Utils::command($fanotify_checker) != 0)
    {
        return 0;
    }

    return 1;
}

sub set_use_fanotify
{
    my $self = shift;
    my $enable = shift;

    my $iniFileName = $self->get_path('.ini');
    my $iniFileContent = KL::Utils::load_ini_file($iniFileName);

    if (!$iniFileContent)
    {
        KL::Utils::warning($MSG_CANNOT_LOAD_INI_FILE);
        return 0;
    }

    $iniFileContent->{'General'}->{'UseFanotify'} = $enable ? 'yes' : 'no';
    if (!KL::Utils::save_ini_file($iniFileName, $iniFileContent))
    {
        KL::Utils::warning($MSG_CANNOT_SAVE_INI_FILE);
        return 0;
    }

    return 1;
}

sub is_argeement_accepted
{
    my $self = shift;
    my $field = shift;
    my $id = shift;

    my $iniFileName = $self->get_path('agreements.ini');
    my $iniFileContent = KL::Utils::load_ini_file($iniFileName);
    if (!$iniFileContent)
    {
        KL::Utils::warning($MSG_CANNOT_LOAD_INI_FILE);
        return 0;
    }

    my $guids = $iniFileContent->{'AcceptedAgreement'}->{$field};

    return index(lc($guids), lc($id)) != -1;
}

sub store_argeement_acceptence
{
    my $self = shift;
    my $field = shift;
    my $guid = shift;

    my $iniFileName = $self->get_path('agreements.ini');
    my $iniFileContent = KL::Utils::load_ini_file($iniFileName);
    if (!$iniFileContent)
    {
        KL::Utils::warning($MSG_CANNOT_LOAD_INI_FILE);
        return 0;
    }

    $iniFileContent->{'AcceptedAgreement'}->{$field} = $guid;

    if (!KL::Utils::save_ini_file($iniFileName, $iniFileContent))
    {
        KL::Utils::warning($MSG_CANNOT_SAVE_INI_FILE);
        return 0;
    }

    return 1;
}

sub disable_cdrom_source_for_debian
{
    my $install_clause = shift;

    my $sh_script = "set -e;  INSTALL_CLAUSE=\"$install_clause\"; " . <<'END_OF_SH_SCRIPT';
uname -a | grep -i "Debian" 1>/dev/null 2>&1 || false
VAR_ASSIGNMENT_CODE=$(apt-config shell ETC_OPT 'Dir::Etc')
eval "${VAR_ASSIGNMENT_CODE}"
VAR_ASSIGNMENT_CODE=$(apt-config shell SOURCELIST_OPT 'Dir::Etc::sourcelist')
eval "${VAR_ASSIGNMENT_CODE}"
SOURCELIST_FILE="/${ETC_OPT}/${SOURCELIST_OPT}"
test -f "${SOURCELIST_FILE}" || false
DEB_CDROM_REGEX='^\s*deb\s+cdrom[ :]'
grep -E "${DEB_CDROM_REGEX}" "${SOURCELIST_FILE}" 1>/dev/null 2>&1 || false
apt-cdrom --no-mount --no-act ident 1>/dev/null 2>&1 && false
TMP_FILE=$(mktemp)
grep -v -E "${DEB_CDROM_REGEX}" "${SOURCELIST_FILE}" > "${TMP_FILE}"
APT_OPTIONS="-o 'Dir::Etc::sourcelist=${TMP_FILE}'"
CLEAN_CODE="; rc=\$?; rm -f '${TMP_FILE}' 2>/dev/null; exit \${rc}"
printf '%s' "${APT_OPTIONS} ${INSTALL_CLAUSE} ${CLEAN_CODE}"
END_OF_SH_SCRIPT

    my $txt=`${sh_script}`;

    return "apt-get " . ($? == 0 ? "${txt}" : "${install_clause}");
}

sub get_build_tools_install_cmd
{
    my $self = shift;
    my $os = KL::Utils::get_os_name();
    my $assume_yes = '-y' if ($self->ai_mode);

    if ($os =~ /Debian/i)
    {
        return disable_cdrom_source_for_debian("${assume_yes} install build-essential");
    }
    elsif ($os =~ /Red Hat/i)
    {
        return "yum ${assume_yes} install gcc make";
    }
    elsif ($os =~ /SUSE/i)
    {
        return "zypper in ${assume_yes} make gcc";
    }


    return "";
}

sub get_kernel_headers_install_cmd
{
    my $self = shift;
    my $os = KL::Utils::get_os_name();
    my $assume_yes = '-y' if ($self->ai_mode);

    if ($os =~ /Debian/i)
    {
        chomp(my $ver = `uname -r`);
        return disable_cdrom_source_for_debian("${assume_yes} install linux-headers-${ver}");
    }
    elsif ($os =~ /Red Hat/i)
    {
        chomp(my $ver = `uname -r`);
        return "yum ${assume_yes} install kernel-headers-${ver} kernel-devel-${ver}";
    }
    elsif ($os =~ /openSUSE/i)
    {
        return "sudo zypper in ${assume_yes} kernel-devel";
    }
    elsif ($os =~ /SUSE/i)
    {
        return "zypper in ${assume_yes} kernel-default-devel";
    }


    return "";
}

sub is_module_loaded
{
    my $self = shift;
    my $module = shift;

    if ($^O =~ /linux/i) {
        my $modules = KL::Utils::read_file('/proc/modules');
        return 1 if $modules =~ /^$module\s/m;
    }

    return 0;
}

sub do_remove_oas_module
{
    my $self = shift;
    my $os_rev  = `uname -r`;
    $os_rev =~ s/\n//;

    my $module = "kav4fs_oas";
    my $module_redir = "redirfs";

    if ($^O =~ /linux/i) {
        system("/sbin/rmmod $module > /dev/null 2>&1") if $self->is_module_loaded($module);
    } else {
        system("kldunload $module");
    }

    my $oasmodule = $self->settings('OAS_KERNELMODULE');

    if (defined $oasmodule) {
        unlink($oasmodule) if -f $oasmodule;
        $self->remove_settings('OAS_KERNELMODULE');
    }

    $oasmodule = $self->settings('OAS_KERNELMODULE_REDIR');
    if (defined $oasmodule) {
        unlink($oasmodule) if -f $oasmodule;
        $self->remove_settings('OAS_KERNELMODULE_REDIR');
    }
}

sub start_oas
{
    my $self = shift;


    my ($ret, $str) = $self->_control_nout('--start-task', $TASK_ID_OAS);

    return $ret == 0;
}

sub start_oas_if_stopped
{
    my $self = shift;
    my ($ret, $str) = $self->_control_nout('--get-task-state', $TASK_ID_OAS);
    return $self->start_oas() if $self->has_matched_line($str,
            '\s*State\s*:\s*Stopped');
    return 0;
}

sub _get_locale_choice_data()
{
    my @supported_locales = ("de_DE.UTF-8", "en_US.UTF-8", "fr_FR.UTF-8", "ja_JP.UTF-8", "ru_RU.UTF-8", "zh_CN.UTF-8");
    @supported_locales = ("en_US.UTF-8", "ru_RU.UTF-8") if ( $LFS_PRODUCT_NAME eq 'kics' );
    my %locale_valid = map {$_ => int(_is_valid_locale($_))} @supported_locales;

    my $current_locale = setlocale("LC_ALL");
    $current_locale =~ s/utf[-]*8/UTF-8/gi;

    my $default_locale = $current_locale if $locale_valid{$current_locale};
    $locale_valid{"C"} = _is_valid_locale("C");

    $default_locale ||= "en_US.UTF-8" if $locale_valid{"en_US.UTF-8"};
    ($default_locale) = grep($locale_valid{$_}, @supported_locales) if !$default_locale;
    $default_locale ||= "C";

    my @locale_options = grep { $default_locale ne $_ } @supported_locales;
    unshift @locale_options, $default_locale;

    my $locale_list = "";
    foreach my $option (@locale_options) {
        $locale_list .= "- $option";
        $locale_list .= " [$MSG_NOT_SUPPORTED_BY_OS]" if (not $locale_valid{$option});
        $locale_list .= "\n";
    }

    return $locale_list, $default_locale, \%locale_valid;
}

sub _is_valid_locale
{
    my $locale = shift;

    my $save_locale = setlocale("LC_ALL");
    my $res = setlocale("LC_ALL", $locale);
    setlocale("LC_ALL", $save_locale);

    return defined($res);
}

sub ai_validate_locale
{
    my $self = shift;
    my $rule = shift;
    my $value = shift;

    if ( _is_valid_locale($value) ) {
        return 1;
    }

    return (0, undef);
}

sub set_service_locale
{
    my $self = shift;
    my $locale = shift;

    $self->settings('LOCALE', $locale);

    KL::Utils::output($MSG_LOCALE_CHANGED, $locale);

    {
        my $ini =  KL::Utils::load_ini_file($self->get_path('.ini'));
        $ini->{General}->{Locale} = $locale;
        KL::Utils::save_ini_file($self->get_path('.ini'), $ini);
    }

    my $restarted = 0;

    if ($restarted) {
        my $fwcontrol = $self->{'fwcontrol'};
        my $appname = $self->appinfo('product', 'title');

        KL::Utils::output($MSG_WAIT_AFTER_RESTART, $appname);

        my $ready = 0;

        for (my $timeout = 0; $timeout < 60; $timeout++) {

            my ($ret, $out) = KL::Utils::execute($fwcontrol, '--get-app-list');

            KL::Utils::debug("$fwcontrol --get-app-list = $ret '$out'");

            if ($ret == 0) {
                if ($out =~ /^product_id:$APP_ID\s.*\sstate:Started$/) {
                    $ready = 1;
                    last;
                }
            }

            sleep(1);
        }

        if (!$ready) {
            KL::Utils::error($MSG_ERROR_NOTREADY_AFTER_RESTART, $appname);
        }
    }

    $self->ai_save('LOCALE', $locale);
}

sub exec_service_comand
{
    my $self   = shift;
    my $comand = shift;

    my $rcd = $self->get_path('init_d_script');
    my $ret = 0;
    my $message;

    if (KL::Utils::is_linux())
    {
        ($ret, $message) = KL::Utils::execute("$rcd", $comand);
    }
    else
    {
        $ret = -1;
        $message = $MSG_NOT_IMPLEMENTED_FOR_NON_LINUX;
    }

    return ($ret, $message)
}

sub generate_and_save_machineid
{
    my $self = shift;

    my $iniFileName = $self->get_path('.ini');
    my $iniFileContent = KL::Utils::load_ini_file($iniFileName);

    if (!$iniFileContent || !$iniFileContent->{'General'}->{'MachineId'})
    {
        my ($ret, $uuid) = KL::Utils::execute('cat', '/sys/class/dmi/id/product_uuid');

        if ($ret != 0)
        {
            ($ret, $uuid) = KL::Utils::execute('cat', '/var/lib/dbus/machine-id');

            my @octets = $uuid =~ /(.{2})/g;

            $uuid = "@octets[0..3]-@octets[4..5]-@octets[6..7]-@octets[8..9]-@octets[10..15]";

            $uuid =~ s/ //g;
        }

        if ($ret != 0)
        {
            ($ret, $uuid) = KL::Utils::execute('cat', '/proc/sys/kernel/random/uuid');
        }

        if ($ret != 0)
        {
            return 0;
        }

        $iniFileContent->{'General'}->{'MachineId'} = lc $uuid;
        if (!KL::Utils::save_ini_file($iniFileName, $iniFileContent))
        {
            KL::Utils::warning($MSG_CANNOT_SAVE_INI_FILE);
            return 0;
        }
    }

    return 1;
}

sub start_service
{
    my $self = shift;

    my $appname = $self->appinfo('product', 'title');
    KL::Utils::output($MSG_TRY_TO_START, $appname);

    my $message;
    my $res;
    ($res, $message) = $self->exec_service_comand("start");

    if ($res != 0)
    {
        KL::Utils::error($message);
        return 0;
    }

    return 1;
}

sub stop_service
{
    my $self = shift;

    my $appname = $self->appinfo('product', 'title');
    KL::Utils::output($MSG_TRY_TO_STOP, $appname);
    my ($ret, $message) = $self->exec_service_comand("stop");

    return $ret == 0;
}

sub restart_service()
{
    my $self = shift;

    my $appname = $self->appinfo('product', 'title');
    KL::Utils::output($MSG_TRY_TO_RESTART, $appname);

    $self->exec_service_comand("stop");

    return $self->exec_service_comand("start");
}

sub check_service_is_running
{
    my $self = shift;

    my ($ret, $message) = $self->exec_service_comand("status");
    return $ret == 0;
}

sub do_setup_service_locale
{
    my $self = shift;

    KL::Utils::header($HEADER_LOCALE);

    my ($choise_list, $default_choice, $choice_valid) = _get_locale_choice_data();

    my $locale;
    while (1) {
        $locale = KL::Utils::ask_question ($default_choice, $MSG_SERVICE_ASK_LOCALE, $choise_list);
        last if $choice_valid->{$locale};
        if (defined($choice_valid->{$locale})) {
            KL::Utils::error($MSG_LOCALE_NOT_SUPPORTED_BY_OS, $locale);
        } else {
            KL::Utils::error($MSG_SERVICE_BAD_LOCALE, $locale);
        }
        print "\n";
    }

    $self->set_service_locale($locale);
}

sub setup_service_locale
{
    my $self = shift;

    my $locale = $self->settings('LOCALE');

    return if defined $locale;

    $self->do_setup_service_locale();
}

sub upgrade_config_files
{
    my $self = shift;

    my $upgrd_ctl = new KL::Upgrade(
        'converter_bin' => $self->get_path('libexec') . "/converter",
        'logger' => $self
    );

    if (!$upgrd_ctl->do_upgrade()) {
        KL::Utils::warning($MSG_UPGRADE_MODULE_WARN, $self->get_applog_path());
        return 0;
    }

    return 1;
}

sub ask_about_ksn
{
    my $self = shift;

    my $agreement_accepted = $self->is_argeement_accepted('KsnaId', 'dfb67b32-a515-d9d7-6730-079cee0d738a');
    if (-f $agreement_accepted)
    {
        return 0;
    }

    KL::Utils::header($MSG_CONFIGURING_KSN);

    my $agreement_template = $self->get_path('ksn_license_template');
    my $agreement_file     = $self->get_sutable_license_file($agreement_template, "ksn");

    my $answ = KL::Utils::ask_boolean('', $MSG_PARTICIPATE_IN_KSN, $agreement_file);

    my $save_value = $answ ? "yes" : "no";
    $self->settings($KSN_ENABLED_SETTING, $save_value);
    $self->ai_save('USE_KSN', $save_value);
}

sub setup_ksn
{
    my $self = shift;

    my $ksn_state = $self->{ 'SETTINGS' }->{''}->{$KSN_ENABLED_SETTING};
    return 1 if (!defined $ksn_state);

    my @settingValue;
    if (lc($ksn_state) eq 'yes')
    {
        push(@settingValue, 'UseKSN=Extended');
        push(@settingValue, '--accept-ksn');
    }
    else
    {
        push(@settingValue, 'UseKSN=No');
    }
    my($ret, $output) = $self->_control_nout('--set-app-settings', @settingValue);

    if ($ret != 0)
    {
        KL::Utils::warning($MSG_FAILED_TO_SETUP_KSN);
        return 0;
    }

    return 1;
}

sub setup_file_interceptor
{
    my $self = shift;

    KL::Utils::header($MSG_CONFIGURING_FILE_INTERCEPTOR);

    KL::Utils::output($MSG_CHECKING_FANOTIFY_AVAILABLE);
    my $useFanotify = $self->check_fanotify_available();
    $self->set_use_fanotify($useFanotify);
    if ($useFanotify)
    {
        KL::Utils::output($MSG_USING_FANOTIFY_INTERCEPTOR);
        return 1;
    }
    else
    {
        KL::Utils::output($MSG_FANOTIFY_NOT_AVAILABLE);
        return $self->do_build_oas_module();
    }
}

sub install_selinux_policy
{
    my $self = shift;
    my $policy_dir = shift;
    my $policy_name = shift;
    my $dir = getcwd;
    my $temp_dir = "${LFS_TEMP_DIRECTORY}selinux/";

    make_path($temp_dir);
    chdir($temp_dir);
    copy("${policy_dir}${policy_name}.te", "${temp_dir}${policy_name}.te");

    my $ret = KL::Utils::compile_selinux_module(${temp_dir}, ${policy_name}) &&
        KL::Utils::install_selinux_module("${temp_dir}${policy_name}.pp") &&
        KL::Utils::enable_selinux_module(${policy_name});

    chdir($dir);
    rmtree($temp_dir);

    return $ret;
}

sub configure_selinux
{
    my ($self, $install_dir) = @_;
    my $type;
    my $default_bin_type = KL::Utils::get_selinux_type_or_default(`command -v "perl"`, "bin_t");
    my $ret = 1;

    if (KL::Utils::can_install_selinux_policy() && $self->install_selinux_policy("${install_dir}${SELINUX_POLICY_DIR}", $LFS_PRODUCT_NAME))
    {
        $type = "${LFS_PRODUCT_NAME}_control_exec_t";
        $ret = KL::Utils::mark_file_with_selinux_type("${install_dir}${BL_FILEPATH}", "${LFS_PRODUCT_NAME}_exec_t") && $ret;
    }
    else
    {
        $type = $default_bin_type;
        $ret = KL::Utils::mark_file_with_selinux_type("${install_dir}${BL_FILEPATH}", ${type}) && $ret;
    }
    $ret = KL::Utils::mark_file_with_selinux_type("${install_dir}${CONTROL_FILEPATH}", ${type}) && $ret;
    $ret = KL::Utils::mark_file_with_selinux_type("${install_dir}${SUPERVISOR_FILEPATH}", ${type}) && $ret;
    if (-x $GUI)
    {
        $ret = KL::Utils::mark_file_with_selinux_type("${install_dir}${GUI}", ${type}) && $ret;
    }
    if (-x $LENA_AUSHAPE_PLUGIN)
    {
        my $default_lib_type = KL::Utils::get_selinux_type_or_default("/lib", "lib_t");
        $ret = KL::Utils::mark_file_with_selinux_type("${install_dir}${LENA_AUSHAPE_PLUGIN}", ${default_bin_type}) && $ret;
        $ret = KL::Utils::mark_file_with_selinux_type("${install_dir}${LENA_AUSHAPE}", ${default_bin_type}) && $ret;
        $ret = KL::Utils::mark_file_with_selinux_type("${install_dir}${LENA_LIBAUSHAPE}", ${default_lib_type}) && $ret;
    }

    if ($ret)
    {
        KL::Utils::output($MSG_CONFIGURE_SELINUX_CONFIGURED);
    }
    return $ret;
}

sub ask_about_selinux
{
    my $self = shift;

    KL::Utils::header($HEADER_CONFIGURE_SELINUX);

    my $answ = KL::Utils::ask_boolean('y', $MSG_CONFIGURE_SELINUX_CONFIRM);
    if (!($answ && $self->configure_selinux()))
    {
        KL::Utils::output($MSG_CONFIGURE_SELINUX_NOT_CONFIGURED);
    }
}

sub ask_about_admin
{
    my $self = shift;

    KL::Utils::header($MSG_GRANTING_ADMIN_ROLE);
    KL::Utils::output($MSG_ADMIN_ROLE_DESCRIPTION);
    print "\n";

    my $username;
    while (1)
    {
        $username = KL::Utils::ask_question('', $MSG_ADMIN_USER_QUESTION);
        if (! $username) {
            last;
        }

        if(! defined(getpwnam($username))) {
            KL::Utils::error($MSG_USER_DOES_NOT_EXIST, $username);
            next;
        }

        last;
    }

    $self->settings('ADMIN_USER', $username);
    $self->ai_save('ADMIN_USER', $username);
}

sub get_scanfile_limit
{
    my $self = shift;
    my ($res, $settings) = $self->get_app_settings_as_simple_hash();

    if ($res != 0)
    {
        return -1;
    }
    if (!defined($settings))
    {
        return -1;
    }

    $settings = $settings->{''};

    return $settings->{'LimitNumberOfScanFileTasks'};
}

sub set_scanfile_limit
{
    my $self = shift;
    my $limit = shift;

    my($ret, $output) = $self->_control_nout('--set-app-settings',
                                             "LimitNumberOfScanFileTasks=$limit");

    if ($ret == 0 && $limit != 0)
    {
        KL::Utils::output("\n<B>" . $MSG_SCANFILE_LIMIT_CHANGED . "</B>", $limit);
    }

    return $ret == 0 ? 1 : 0
}

sub find_gui_session_logind
{
    my $listSessionsCmd = "env LC_ALL=C loginctl --no-legend list-sessions 2>/dev/null";
    my $listSessions;
    if (!defined(open($listSessions, "-|", $listSessionsCmd)))
    {
        KL::Utils::warning($MSG_CANNOT_FORK);
        return;
    }

    my @sessions;
    foreach (<$listSessions>)
    {
        chomp;
        s/^\s+|\s+$//g;
        my ($session, $uid, $user, $seat, $tail) = split(/\s+/, $_, 5);
        if ($seat eq "seat0")
        {
            push(@sessions, {session=>$session, user=>$user});
        }
    }

    if (!close($listSessions))
    {
        return;
    }

    foreach my $s (@sessions)
    {
        my $showSessionCmd = "env LC_ALL=C loginctl show-session $s->{session} 2>/dev/null";
        my $showSession;

        if (!defined(open($showSession, "-|", $showSessionCmd)))
        {
            KL::Utils::warning($MSG_CANNOT_FORK);
            return;
        }

        foreach (<$showSession>)
        {
            chomp;
            my ($key, $value) = split('=', $_, 2);
            if ($key eq "Type")
            {
                $s->{type} = $value;
            }
            elsif ($key eq "Class")
            {
                $s->{class} = $value;
            }
            elsif ($key eq "Leader")
            {
                $s->{leader} = $value;
            }
        }

        if (!close($showSession))
        {
            return;
        }

        if (($s->{type} eq "x11" || $s->{type} eq "wayland") && $s->{class} eq "user")
        {
            return ($s->{user}, $s->{leader});
        }
    }

    return;
}

sub find_gui_session_utmp
{
    my $whoCmd = "env LC_ALL=C who -u 2>/dev/null";
    my $who;
    if (!defined(open($who, "-|", $whoCmd)))
    {
        KL::Utils::warning($MSG_CANNOT_FORK);
        return;
    }

    my @sessions;
    foreach (<$who>)
    {
        chomp;
        s/^\s+|\s+$//g;
        my ($name, $line, $mon, $day, $time, $idle, $pid, $comment) = split(/\s+/, $_, 8);
        if ($line eq ":0" || ($line =~ /^tty/ && $comment eq "(:0)"))
        {
            push(@sessions, {name=>$name, pid=>$pid});
        }
    }

    if (!close($who))
    {
        KL::Utils::warning($MSG_CANNOT_EXECUTE, $whoCmd, $!);
        return;
    }

    if (scalar(@sessions) == 0)
    {
        return;
    }

    if (system('/usr/bin/getent', 'passwd', $sessions[0]->{name}) != 0)
    {
        return;
    }

    return ($sessions[0]->{name}, $sessions[0]->{pid});
}

sub ps
{
    my ($psCmd) = @_;
    my $ps;
    if (!open($ps, "-|", $psCmd))
    {
        KL::Utils::warning($MSG_CANNOT_FORK);
        return;
    }

    my @pids;
    foreach (<$ps>)
    {
        chomp;
        s/^\s+|\s+$//g;
        push(@pids, $_);
    }

    if (!close($ps))
    {
        KL::Utils::warning($MSG_CANNOT_EXECUTE, $psCmd, $!);
        return;
    }

    return @pids;
}

sub get_gui_env
{
    my $pid = shift;
    my $environFile = "/proc/$pid/environ";
    my $environ;
    if (!open($environ, "<", $environFile))
    {
        KL::Utils::warning($MSG_CANNOT_OPEN, $environFile, $!);
        return;
    }

    my %env;
    my $sep = $/;
    $/ = "\0";
    foreach (<$environ>)
    {
       chomp;
       my ($name, $value) = split("=", $_, 2);
       $env{$name} = $value;
    }
    $/ = $sep;
    close($environ);

    return \%env;
}

sub get_gui_env_recursive
{
    my ($psCmd) = @_;

    my @pids = ps($psCmd);

    foreach (@pids)
    {
        my $env = get_gui_env($_);
        if (defined($env) &&
            defined($env->{DISPLAY}) &&
            defined($env->{DBUS_SESSION_BUS_ADDRESS}))
        {
            return $env;
        }
    }

    foreach(@pids)
    {
        my $psCmd = "ps -o pid= --ppid $_";
        my $env = get_gui_env_recursive($_);
        if (defined($env) &&
            defined($env->{DISPLAY}) &&
            defined($env->{DBUS_SESSION_BUS_ADDRESS}))
        {
            return $env;
        }
    }

    return;
}

sub start_gui
{
    KL::Utils::output($MSG_GUI_LOOKING_FOR_SESSION);

    my ($sessionUser, $sessionLeaderPid) = find_gui_session_logind();
    if (!defined($sessionLeaderPid))
    {
        ($sessionUser, $sessionLeaderPid) = find_gui_session_utmp();
    }

    if (!defined($sessionLeaderPid))
    {
        KL::Utils::output($MSG_GUI_NO_SESSION_FOUND);
        return 0;
    }

    KL::Utils::output($MSG_GUI_SESSION_FOUND, $sessionUser, $sessionLeaderPid);

    KL::Utils::output($MSG_GUI_GETTING_ENVIRONMENT);

    my $env = get_gui_env_recursive("ps -o pid= -p $sessionLeaderPid -u $sessionUser");

    if (!defined($env))
    {
        KL::Utils::warning($MSG_GUI_CANNOT_GET_ENVIRONMENT);
        return 0;
    }

    KL::Utils::output($MSG_GUI_GETTING_USER_INFO);
    my ($uid, $gid) = (getpwnam($sessionUser))[2,3];
    if (!defined($uid) || !defined($gid))
    {
        KL::Utils::warning($MSG_CANNOT_GETPWNAM, $sessionUser, $!);
        return 0;
    }
    KL::Utils::output($MSG_GUI_GOT_USER_INFO, $uid, $gid);

    KL::Utils::output($MSG_GUI_STARTING);
    my $pid;
    $pid = fork();
    if (!defined($pid))
    {
        KL::Utils::warning($MSG_CANNOT_FORK);
        return 0;
    }
    if ($pid)
    {
        return 1;
    }

    POSIX::setsid();
    POSIX::setgid($gid);
    POSIX::setuid($uid);
    $pid = fork();
    if (!defined($pid) || $pid)
    {
        exit;
    }

    chdir("/");
    delete(@ENV{keys(%ENV)});
    @ENV{keys(%$env)} = @$env{keys(%$env)};
    open(STDIN, "<", "/dev/null");
    open(STDOUT, ">", "/dev/null");
    open(STDERR, ">", "/dev/null");
    exec('/usr/bin/env', "FONTCONFIG_FILE=$LFS_FONTS_CONF", $GUI, '--tray-only');
}

sub setup_gui
{
    my $self = shift;
    my $reconfigure = shift;

    unless (-x $GUI)
    {
        if ($reconfigure)
        {
            $self->set_scanfile_limit(0);
        }
        return 1;
    }

    KL::Utils::header($MSG_CONFIGURING_GUI);

    $self->start_gui();

    if ($self->get_scanfile_limit() == 0)
    {
        $self->set_scanfile_limit(5);
    }

    return 1;
}

sub grant_admin_role
{
    my $self = shift;

    my $username = $self->settings('ADMIN_USER');
    if (!$username)
    {
        return (0, '');
    }

    return $self->_control_nout('--grant-role', 'admin', $username);
}


sub do_auto_install
{
    my $self = shift;
    my $warnings = 0;
    my $fatal_errors = 0;

    my ($opt, $res, $message);

    my $reconfigure = 0;

    $opt = $self->ai_get_value('EULA_AGREED');
    if (lc($opt) ne 'yes') {
        KL::Utils::error('EULA is not agreed');
        return 0;
    }

    $opt = $self->ai_get_value('PRIVACY_POLICY_AGREED');
    if (lc($opt) ne 'yes') {
        KL::Utils::error('Privacy Policy is not agreed');
        return 0;
    }

    my $eula = $self->get_sutable_license_file($self->get_path('license_template'), "eula");
    $self->store_argeement_acceptence('EulaId', '18cbd897-05cc-535f-da7e-b570b4795b3e');
    $self->settings('EULA_AGREED', 'Yes');

    {
        $self->generate_and_save_machineid();

        if ($self->check_service_is_running())
        {
            if (!$self->stop_service())
            {
                KL::Utils::warning($MSG_CANNOT_STOP_SERVICE);
                return 0;
            }
            $reconfigure = 1;
        }

        {
            my $serviceLocale = $self->ai_get_value('LOCALE');
            if (not defined($serviceLocale) or $serviceLocale eq '')
            {
                $serviceLocale = setlocale("LC_ALL");
            }
            $self->set_service_locale($serviceLocale);
        }

        {
            if (!$self->setup_file_interceptor())
            {
                $warnings++;
            }
        }
    }

    {
        $opt = $self->ai_get_value('USE_KSN');
        $self->settings($KSN_ENABLED_SETTING, $opt);
    }

    {
        my $adm_user = $self->ai_get_value('ADMIN_USER');
        $self->settings('ADMIN_USER', $adm_user);
    }

    if (KL::Utils::is_selinux_present())
    {
        KL::Utils::header($HEADER_CONFIGURE_SELINUX);

        my $selinux_confirm = $self->ai_get_value('CONFIGURE_SELINUX');
        $self->settings('CONFIGURE_SELINUX', $selinux_confirm);
        if (!(lc($selinux_confirm) eq 'yes' && $self->configure_selinux())) {
            KL::Utils::output($MSG_CONFIGURE_SELINUX_NOT_CONFIGURED);
        }
    }

    if (!$self->start_service())
    {
        return 0;
    }

    {
        my ($res, $message) = $self->grant_admin_role();
        if ($res != 0)
        {
            KL::Utils::warning($message);
            $warnings++;
        }
    }

    {
        if (!$self->setup_ksn())
        {
            $warnings++;
        }

        my $update_source = $self->ai_get_value('UPDATER_SOURCE');
        if (defined $update_source)
        {
            my ($res, $message) = $self->set_updater_source($update_source);
            if ($res != 0)
            {
                KL::Utils::warning($message);
                $warnings++;
            }
        }
    }

    {
        my $proxy_server = $self->ai_get_value('PROXY_SERVER');
        if (defined $proxy_server)
        {
            my ($res, $message) = $self->set_http_proxy($proxy_server);
            if ($res != 0)
            {
                KL::Utils::warning($message);
                $warnings++;
            }
        }
    }

    my $run_update =  $self->ai_get_value('UPDATE_EXECUTE');
    if (lc($run_update) eq 'yes')
    {
        if($self->download_bases() != 0)
        {
            KL::Utils::warning($MSG_FAIL_UPDATE_MESSAGE);
            $warnings++;
        }
    }


    KL::Utils::header($HEADER_INSTALL_KEY);

    my $activation = $self->ai_get_value('INSTALL_LICENSE');
    if (!$self->do_install_activation($activation))
    {
        $warnings++;
    }

    {
        if (!$self->setup_gui($reconfigure))
        {
            ++$warnings;
        }
    }

    $self->start_oas();

    $self->check_klnagent_running();

    if ($warnings)
    {
        $self->{_AI_WARNINGS} = $warnings;
    }

    return 1;
}

sub do_postinstall
{
    my $self = shift;

    my $reconfigure = 0;

    $self->generate_and_save_machineid();

    if ($self->check_service_is_running())
    {
        if (KL::Utils::ask_boolean('n', $MSG_STOP_SERVICE_TO_PROCEED))
        {
            if (!$self->stop_service())
            {
                KL::Utils::warning($MSG_CANNOT_STOP_SERVICE);
                return 0;
            }
            $reconfigure = 1;
        }
        else
        {
            KL::Utils::output($MSG_CONFIGURATION_ABORTED);
            return 0;
        }
    }

    $self->setup_service_locale();

    my $res = $self->show_eula_agreement();
    if ($res != 0)
    {
        return 1;
    }

    $self->ask_about_ksn();
    $self->ask_about_admin();
    $self->setup_file_interceptor();
    if (KL::Utils::is_selinux_present())
    {
        $self->ask_about_selinux();
    }

    if (!$self->start_service())
    {
        return 0;
    }

    $self->grant_admin_role();
    $self->setup_ksn();

    if ($self->check_klnagent_running())
    {
        $self->setup_gui($reconfigure);

        KL::Utils::output($MSG_SETUP_SCRIPT_COMPLETED_SUCCESSFULLY);
        KL::Utils::warning($MSG_KLNAGENT_IS_RUNNING_ON_THE_COMPUTER);
    }
    else
    {
        my $ask_update_settings = 1;
        if (!$self->update_bases($ask_update_settings))
        {
            return 0;
        }

        $self->setup_autoupdates();

        if (!$self->check_license())
        {
            return 0;
        }

        $self->setup_gui($reconfigure);
        $self->start_oas();
    }

    return 1;
}

sub do_uninstall
{
    my $self = shift;

    $self->do_remove_oas_module();

    return 1;
}

sub update_bases()
{
    my $self = shift;
    my $ask_update_settings = shift;
    my $set_update_settings = 1;
    my $enable_patches = 0;

    if (0 != $self->set_update_patch_mode($enable_patches))
    {
        return 0;
    }

    my $res = $self->do_update_bases($ask_update_settings, $set_update_settings);

    $enable_patches = 1;
    $self->set_update_patch_mode($enable_patches);

    if ($res)
    {
        $res = $self->restart_if_needed();
    }

    return $res;
}

sub do_update_bases()
{
    my $self = shift;
    my $ask_update_settings = shift;
    my $set_update_settings = 1;
    my $skip_run_update = 0;

    while ($set_update_settings)
    {
        if ($ask_update_settings)
        {
            my $res = $self->setup_update_source($TASK_ID_UPDATER);
            if($res != 0)
            {
                if ($res < 0)
                {
                    $skip_run_update = 1;
                }
                else
                {
                    return 0;
                }
            }

            if(0 != $self->setup_proxy())
            {
                return 0;
            }
        }

        $set_update_settings = 0;
        if (!$skip_run_update && KL::Utils::ask_boolean('y', $MSG_RUN_UPDATER))
        {
            $self->ai_save('UPDATE_EXECUTE', 'Yes');
            my $update_result = $self->download_bases();
            if ($update_result != 0)
            {
                KL::Utils::warning($MSG_FAIL_UPDATE_MESSAGE);
            }

            my ($res, $has_bases) = $self->has_bases();
            if ($res != 0)
            {
                KL::Utils::warning($has_bases);
                return 0;
            }
            if ($has_bases)
            {
                if ($update_result == 0)
                {
                    KL::Utils::output($MSG_UPDATE_OK_BASES_ARE_DOWNLOADED);
                }
                else
                {
                    KL::Utils::warning($MSG_UPDATE_FAIL_BUT_PRODUCT_HAS_VALID_BASES);
                }
            }
            else
            {
                $set_update_settings =
                    KL::Utils::ask_boolean('y', $MSG_SET_UPDATE_SETTINGS_ONE_MORE);
            }
        }
    }

    return 1;
}

sub set_update_patch_mode()
{
    my $self = shift;
    my $enable = shift;

    my @options;

    if ($enable)
    {
        push @options, 'ApplicationUpdateMode', 'DownloadOnly';
    }
    else
    {
        push @options, 'ApplicationUpdateMode', 'Disabled';
    }

    return $self->refresh_task_settings($TASK_ID_UPDATER, @options);
}

sub restart_if_needed()
{
    my $self = shift;
    my ($res, $str) = $self->_control_nout('--app-info');

    if ($res == $RC_RESTART_REQUIRED)
    {
        my $message;
        ($res, $message) = $self->restart_service();

        if ($res != 0)
        {
            KL::Utils::error($message);
            return 0;
        }
    }

    return 1;
}

sub check_license()
{
    my $self = shift;

    KL::Utils::header($HEADER_INSTALL_KEY);

    my ($lq_res, $lic_request) = $self->_control_nout('-L', '--query');

    if ($lq_res == 64)
    {
        return 0;
    }

    my $use_curr_lic = 0;
    my $is_license_present = ($lq_res == 0);

    if ($is_license_present)
    {
        my $is_commercial =
            ($lic_request =~ /^\s*License type\s*:\s*Commercial\s*$/im);

        my $is_expired =
            ($lic_request =~ /^\s*Days remaining until expiration\s*:\s*0\s*$/im);

        my $is_blacklisted =
            ($lic_request =~ /^\s*License status\s*:\s*Blacklisted\s*$/im);

        if ($is_commercial || !$is_expired)
        {
            my $question;

            if ($is_expired)
            {
                $question = KL::Utils::translate_string(
                    $MSG_INVALID_LICENSE_IS_ALREADY_INSTALLED,
                    $is_commercial ? $MSG_COMMERCIAL_LICENSE
                                   : $MSG_TRIAL_LICENSE,
                    $is_blacklisted ? $MSG_INVALID_LICENSE
                                    : $MSG_OUTDATED_LICENSE);
            }
            else
            {
                $question = KL::Utils::translate_string(
                    $MSG_VALID_LICENSE_IS_ALREADY_INSTALLED,
                    $is_commercial ? $MSG_COMMERCIAL_LICENSE
                                   : $MSG_TRIAL_LICENSE);
            }

            $use_curr_lic =
                KL::Utils::ask_boolean($is_expired ? 'n' : 'y', $question);
            if (not $use_curr_lic)
            {
                $self->_control('--remove-active-key');
            }
        }
    }


    if (not $use_curr_lic)
    {
        $self->install_activation($is_license_present);
    }

    return 1;
}

1;
