#include "klflt.h"

#include <linux/kthread.h>
#include <linux/list.h>
#include <linux/namei.h>
#include <linux/wait.h>
#include <linux/errno.h>

#define INTERCEPT_START_TIMEOUT 20000
#define INTERCEPT_STOP_TIMEOUT 20000
static atomic_t intercept_started = ATOMIC_INIT(0);
static atomic_t intercept_aborted = ATOMIC_INIT(0);
static DECLARE_WAIT_QUEUE_HEAD(intercept_waitq);
static DECLARE_WAIT_QUEUE_HEAD(intercept_start_waitq);
static DEFINE_MUTEX(intercept_mutex);
static struct task_struct *intercept_thrd = NULL;

static int klflt_intercept_stop(void);

static struct klflt_mounts_ops mount_ops = {
    .add_mount = NULL,
    .remove_mount = NULL
};

static void klflt_intercept_abort(void)
{
    pr_debug("%s\n", __func__);

    atomic_xchg(&intercept_aborted, 1);
    wake_up(&intercept_waitq);
}

static struct klflt_watchdog_ops watchdog_ops = {
    .invoke = klflt_intercept_abort,
};

extern klflt_module_config_t klflt_module_config;

static int klflt_intercept_thread(void* unused)
{
    int rv = 0;
    pr_debug("%s started\n", __func__);

    atomic_xchg(&intercept_aborted, 0);

    rv = klflt_rfs_init(&mount_ops);
    if (rv < 0)
    {
        pr_err("klflt: failed to initialize RFS system. err=%d\n", rv);
        goto error_rfs;
    }

    rv = klflt_smb_init(&mount_ops);
    if (rv < 0)
    {
        pr_err("klflt: failed to initialize smb subsystem\n");
        goto error_smb;
    }

    rv = klflt_mounts_handler_init(&mount_ops);
    if (rv < 0)
    {
        pr_err("klflt: failed to initialize mounts monitoring subsystem\n");
        goto error_mounts;
    }

    rv = klflt_cache_start(klflt_module_config.cache_size);
    if (rv < 0)
    {
        pr_err("klflt: failed to initialize cache\n");
        goto error_cache;
    }

    rv = klflt_check_init();
    if (rv < 0)
    {
        pr_err("klflt: failed to initialize check module. err=%d\n", rv);
        goto error_check;
    }

    rv = klflt_watchdog_init(&watchdog_ops);
    if (rv < 0)
    {
        pr_err("klflt: failed to initialize klflt watchdog. err=%d\n", rv);
        goto error_watchdog;
    }
    atomic_xchg(&intercept_started, 1);
    wake_up(&intercept_start_waitq);

    while (!kthread_should_stop())
    {

        wait_event_interruptible(intercept_waitq,
            atomic_read(&intercept_aborted) || kthread_should_stop());
        if (atomic_read(&intercept_aborted))
        {
            pr_err("klflt: file interceptor stopped due to timeout\n");
            break;
        }
    }

    klflt_watchdog_destroy();
error_watchdog:
    klflt_check_destroy();
error_check:
    klflt_cache_stop();
error_cache:
    klflt_mounts_handler_destroy();
error_mounts:
    klflt_smb_destroy();
error_smb:
    klflt_rfs_destroy();
error_rfs:
    atomic_xchg(&intercept_started, 0);
    wake_up(&intercept_start_waitq);

    pr_debug("%s finished\n", __func__);
    return rv;
}

static int klflt_intercept_start(void)
{
    int rv = 0;

    pr_debug("%s\n", __func__);

    mutex_lock(&intercept_mutex);
    if (atomic_read(&intercept_started))
    {
        pr_err("%s: already started\n", __func__);
        rv = -EALREADY;
        goto exit;
    }

    intercept_thrd = kthread_run(klflt_intercept_thread, NULL, "klflt_intercept");
    if (IS_ERR(intercept_thrd))
    {
        rv = PTR_ERR(intercept_thrd);
        pr_err("klflt: filed to start klflt_intercept thread, err=%d\n", rv);
        goto exit;
    }

    if (!wait_event_interruptible_timeout(intercept_start_waitq,
            atomic_read(&intercept_started),
            msecs_to_jiffies(INTERCEPT_START_TIMEOUT)))
    {
        pr_err("%s: Intercept start timeout expired", __func__);
        kthread_stop(intercept_thrd);
        rv = -ETIMEDOUT;
    }

exit:
    mutex_unlock(&intercept_mutex);
    return rv;
}

int klflt_intercept_stop(void)
{
    int rv = 0;

    pr_debug("%s\n", __func__);

    mutex_lock(&intercept_mutex);
    if (!atomic_read(&intercept_started)) {
        pr_err("%s: already stopped\n", __func__);
        goto exit;
    }

    rv = kthread_stop(intercept_thrd);
    if (rv)
    {
        pr_err("klflt: failed to stop klflt_intercept thread, err=%d\n", rv);
    }

   if (!wait_event_interruptible_timeout(intercept_start_waitq,
            !atomic_read(&intercept_started),
            msecs_to_jiffies(INTERCEPT_START_TIMEOUT)))
    {
        pr_warn("%s: Intercept stop timeout expired", __func__);
        rv = -ETIMEDOUT;
    }

exit:
    mutex_unlock(&intercept_mutex);
    return rv;
}

static struct klflt_interceptor_ops intercept = {
    .start = klflt_intercept_start,
    .stop  = klflt_intercept_stop
};

int klflt_intercept_init(struct klflt_interceptor_ops* interceptor_ops)
{
    pr_debug("%s\n", __func__);

    *interceptor_ops = intercept;

    return 0;
}

void klflt_intercept_destroy(void)
{
    pr_debug("%s\n", __func__);

    klflt_intercept_stop();
}